// TYPEDEF(type_id,	pac_type,	c_type,		size)
TYPE_DEF(INT8, 		"int8",		"int8",		1)
TYPE_DEF(INT16,		"int16",	"int16",	2)
TYPE_DEF(INT32,	 	"int32",	"int32",	4)
TYPE_DEF(INT64,	 	"int64",	"int64",	8)
TYPE_DEF(UINT8,		"uint8",	"uint8",	1)
TYPE_DEF(UINT16,	"uint16",	"uint16",	2)
TYPE_DEF(UINT32, 	"uint32",	"uint32",	4)
TYPE_DEF(UINT64, 	"uint64",	"uint64",	8)
TYPE_DEF(EMPTY,		"empty",	"", 		0)
